###############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import division
from __future__ import print_function


def le(h):
	"""
	Little-endian, takes a 16b number and returns an array arrange in little
	endian or [low_byte, high_byte].
	"""
	h &= 0xffff  # make sure it is 16 bits
	return [h & 0xff, h >> 8]


def angle2int(angle, degrees=True):
	if degrees:
		if not (0.0 <= angle <= 300.0):
			raise Exception('angle2int() [deg] out of bounds: {}'.format(angle))
		angle = le(int(angle/300.0*1023))  # degrees
	else:
		if not (0.0 <= angle <= 5.23598776):
			raise Exception('angle2int() [rads] out of bounds: {}'.format(angle))
		angle = le(int(angle/5.23598776*1023))  # radians
	return angle



# ##############################################
# # The MIT License (MIT)
# # Copyright (c) 2016 Kevin Walchko
# # see LICENSE for full details
# ##############################################
#


from __future__ import division
from __future__ import print_function


class Debug(object):
	"""
	Not sure about the value of this ...
	"""

	# def packetToDict(pkt):
	# 	"""
	# 	Given a packet, this turns it into a dictionary ... is this useful?
    #
	# 	in: packet, array of numbers
	# 	out: dictionary (key, value)
	# 	"""
    #
	# 	d = {
	# 		'id': pkt[4],
	# 		# 'instruction': xl320.InstrToStr[pkt[7]],
	# 		# 'length': (pkt[6] << 8) + pkt[5],
	# 		# 'params': pkt[8:-2],
	# 		'Model Number': (pkt[10] << 8) + pkt[9],
	# 		'Firmware Ver': pkt[11],
	# 		# 'Error': ErrorStatusMsg[pkt[8]],
	# 		# 'crc': pkt[-2:]
	# 	}
    #
	# 	return d

	def hex_decode(self, data):
		"""
		Takes an array of number and turns them into a string of hex for printing.

		in: data - array of numbers [1,2,33,42,234]
		out: str - '0x1 0x02 0x21 0x2a 0xea'
		"""
		return ''.join(map('0x{:02X} '.format, data))

	def prettyPrintPacket(self, ctrl_table):
		"""
		This will pretty print out a packet's fields.

		in: dictionary of a packet
		out: nothing ... everything is printed to screen
		"""
		print('---------------------------------------')
		print("{:.<29} {}".format('id', ctrl_table['id']))
		ctrl_table.pop('id')
		for key, value in ctrl_table.items():
			print("{:.<29} {}".format(key, value))



##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import division
from __future__ import print_function
from .utils import le


crc_table = [
	0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
	0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
	0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
	0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
	0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
	0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
	0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
	0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
	0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
	0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
	0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
	0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
	0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
	0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
	0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
	0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101,
	0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
	0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
	0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
	0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
	0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
	0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
	0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
	0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
	0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
	0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
	0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
	0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
	0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
	0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
	0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
	0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202
]


class XL320(object):
	SERVO_ID = 2

	# --------- INSTRUCTIONS -----
	PING      = 0x01
	READ      = 0x02
	WRITE     = 0x03
	REG_WRITE = 0x04
	ACTION    = 0x05
	RESET     = 0x06
	REBOOT    = 0x08
	STATUS    = 0x55
	SYNC_READ  = 0x82
	SYNC_WRITE = 0x83
	BULK_READ  = 0x92
	BULK_WRITE = 0x93

	# -------- EEPROM -------------
	MODEL_NUMBER    = 0
	VER_FIRMWARE    = 2
	ID              = 3
	BAUD_RATE       = 4
	DELAY_TIME      = 5
	CW_ANGLE_LIMIT  = 6   # min angle, default 0
	CCW_ANGLE_LIMIT = 8   # max angle, default 300
	CONTROL_MODE    = 11  # joint or wheel mode, default joint (servo)
	MAX_TORQUE      = 15
	RETURN_LEVEL    = 17

	# -------- RAM ----------------
	TORQUE_ENABLE    = 24  # servo mode on/off - turn into wheel
	LED              = 25
	GOAL_POSITION    = 30
	GOAL_VELOCITY    = 32
	GOAL_TORQUE      = 35
	PRESENT_POSITION = 37  # current servo angle
	PRESENT_SPEED    = 39  # current speed
	PESENT_LOAD      = 41  # current load
	PESENT_VOLTAGE   = 45  # current voltage
	PESENT_TEMP      = 46  # current temperature
	MOVING           = 49
	HW_ERROR_STATUS  = 50
	PUNCH            = 51

	# --------- OTHER -------------
	RESET_ALL                  = 0xFF
	RESET_ALL_BUT_ID           = 0x01
	RESET_ALL_BUT_ID_BAUD_RATE = 0x02
	LED_WHITE                  = 7
	LED_BLUE_GREEN             = 6
	LED_PINK                   = 5
	LED_BLUE                   = 4
	LED_YELLOW                 = 3
	LED_GREEN                  = 2
	LED_RED                    = 1
	LED_OFF                    = 0
	BROADCAST_ADDR             = 0xfe  # a packet with this ID will go to all servos
	WHEEL_MODE                 = 1
	JOINT_MODE                 = 2  # normal servo
	DR_9600                    = 0  # 0: 9600, 1:57600, 2:115200, 3:1Mbps
	DR_57600                   = 1
	DR_115200                  = 2
	DR_1000000                 = 3

	def __init__(self):
		# print('<<< WARNING: this is not setup yet >>>')
		pass

	def check_sum(self, data_blk):
		"""
		Calculate crc

		in: data_blk - entire packet except last 2 crc bytes
		out: crc_accum - 16 word
		"""
		data_blk_size = len(data_blk)
		crc_accum = 0
		for j in range(data_blk_size):
			i = ((crc_accum >> 8) ^ data_blk[j]) & 0xFF
			crc_accum = ((crc_accum << 8) ^ crc_table[i])
			crc_accum &= 0xffff  # keep to 16 bits

		return crc_accum

	def makePacket(self, ID, instr, params=None):
		"""
		This makes a generic packet.

		TODO: look a struct ... does that add value using it?

		0xFF, 0xFF, 0xFD, 0x00, ID, LEN_L, LEN_H, INST, PARAM 1, PARAM 2, ..., PARAM N, CRC_L, CRC_H]
		in:
			ID - servo id
			instr - instruction
			params - [register, instruction parameter values]
		out: packet
		"""
		pkt = []
		# [header, reserved, 0x00, ID, len low, len high, instruction]
		pkt += [0xFF, 0xFF, 0xFD, 0x00, ID, 0x00, 0x00, instr]  # header
		# pkt += [0x00]  # reserved byte
		# pkt += [ID]
		# pkt += [0x00, 0x00]  # length placeholder
		# pkt += [instr]  # instruction
		# if reg:
		# 	pkt += le(reg)  # not everything has a register
		if params:
			pkt += params    # not everything has parameters

		length = le(len(pkt) - 5)  # length = len(packet) - (header(3), reserve(1), id(1))
		pkt[5] = length[0]  # L
		pkt[6] = length[1]  # H

		crc = self.check_sum(pkt)
		pkt += le(crc)

		return pkt

	def find_packets(self, pkt):
		"""
		Search through a string of binary for a valid xl320 package.

		in: buffer to search through
		out: a list of valid data packet
		"""
		# print('findpkt', pkt)
		# print('-----------------------')
		ret = []
		while len(pkt)-10 >= 0:
			if pkt[0:4] != [0xFF, 0xFF, 0xFD, 0x00]:
				pkt.pop(0)  # get rid of the first index
				# print(' - pop:', pkt)
				continue
			# print(' > good packet')
			length = (pkt[6] << 8) + pkt[5]
			# print(' > length', length)
			crc_pos = 5 + length
			pkt_crc = pkt[crc_pos:crc_pos + 2]
			crc = le(self.check_sum(pkt[:crc_pos]))
			# if len(pkt) < (crc_pos + 1):
			# 	print('<<< need more data for findPkt >>>')

			# print(' > calc crc', crc)
			# print(' > pkt crc', pkt_crc)
			if pkt_crc == crc:
				pkt_end = crc_pos+2
				ret.append(pkt[:pkt_end])
				# print(' > found:', pkt[:pkt_end])
				# print(' > pkt size', pkt_end)
				del pkt[:pkt_end]
				# print(' > remaining:', pkt)
			else:
				pkt_end = crc_pos+2
				# print(' - crap:', pkt[:pkt_end])
				del pkt[:pkt_end]
		# print('findpkt ret:', ret)
		return ret

	def status_packet(self, pkt):
		def getError(err):
			errors = [
				'Input Voltage',  # 0
				'Angle Limit',
				'Overheating',
				'Range',
				'Checksum',
				'Overload',
				'Instrunction',
				'None'  # 7

			]
			ret = None
			if err != 128:
				err_str = []
				for i in range(0, 8):
					if (err >> i) & 1:
						err_str.append(errors[i])
				ret = ','.join(err_str)
			else:
				ret = errors[7]
			return ret

		ret = {
			'id': pkt[2],
			'error str': getError(pkt[4]),
			'error num': pkt[4],
			'params': pkt[5:-1],
			'raw': list(pkt)
		}
		return ret



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################
# Send ping commands to all servos

from __future__ import print_function
from pyservos import Packet
from pyservos import ServoSerial
import sys
import argparse
import time
import pyservos


valid_return = False
found_servos = {}


def print_status_pkt(info):
	print('---------------------------------------')
	print("{:.<29} {}".format('id', info['id']))
	print("{:.<29} {}".format('Error', info['error str']))
	print('raw pkt: {}'.format(info['raw']))


def ping(port, rate, ID, retry=3):
	"""
	Sends a ping packet to ID's from 0 to maximum and prints out any returned
	messages.

	Actually send a broadcast and will retry (resend) the ping 3 times ...
	"""
	global valid_return

	s = ServoSerial(port, rate)

	if ID < 0:
		print('Pinging ALL servos')
		ID = None
	else:
		print('Pinging servo:', ID)

	try:
		s.open()
	except Exception as e:
		# print('Error opening serial port:')
		print('-'*40)
		print(sys.argv[0], ':')
		print(e)
		exit(1)

	servo = Packet(pyservos.AX12)

	pkt = servo.makePingPacket(ID)
	# print('ping', pkt)
	s.write(pkt)

	# as more servos add up, I might need to increase the cnt number???
	for cnt in range(retry):
		ans = s.read()

		if ans:
			valid_return = True
			pkts = servo.decodePacket(ans)
			for pkt in pkts:
				info = servo.processStatusPacket(pkt)
				if info['id'] not in found_servos.keys():
					found_servos[info['id']] = info
				# print_status_pkt(info, pkt)
		else:
			if not valid_return:
				print('Try {}: no servos found'.format(cnt))

		time.sleep(0.1)

	keys = found_servos.keys()
	keys.sort()

	for key in keys:
		print_status_pkt(found_servos[key])

	s.close()


DESCRIPTION = """
Sends out a ping packet and prints all of the returned packets. If you don't
specify a specific id number, the program defaults to 'all'.

Example:

kevin@logan bin $ ./servo_ping.py /dev/tty.usbserial-AL034G1T
Finding all servos:
Opened /dev/tty.usbserial-AL034G1T @ 1000000
Try 0: no servos found
---------------------------------------
id........................... 1
Error........................ None
raw pkt: [255, 255, 1, 2, 0, 252]
"""


def handleArgs():
	parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
	parser.add_argument('-r', '--rate', help='servo baud rate', type=int, default=1000000)
	parser.add_argument('-i', '--id', help='ping servo ID', type=int, default=-1)
	parser.add_argument('port', help='serial port name, set to "dummy" for testing', type=str)

	args = vars(parser.parse_args())
	return args


if __name__ == '__main__':
	args = handleArgs()
	ping(port=args['port'], rate=args['rate'], ID=args['id'])



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function
from pyservos import Packet
from pyservos import ServoSerial
import pyservos
import argparse


DESCRIPTION = """
Set the angle of a servo in degrees.

Example: set servo 3 to angle 45

./set_angle /dev/serial0 45 -i 3
"""


def handleArgs():
	parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
	parser.add_argument('-i', '--id', help='servo id', type=int, default=0)
	parser.add_argument('port', help='serial port  or \'dummy\' for testing', type=str)
	parser.add_argument('angle', help='servo angle in degrees: 0.0 - 300.0', type=float)

	args = vars(parser.parse_args())
	return args


def main():
	args = handleArgs()

	ID = args['id']
	port = args['port']
	angle = args['angle']

	print('Setting servo[{}] to {:.2f} on port {}'.format(ID, angle, port))

	serial = ServoSerial(port=port)
	serial.open()

	servo = Packet(pyservos.AX12)

	pkt = servo.makeServoMovePacket(ID, angle)
	ans = serial.sendPkt(pkt)  # send packet to servo
	if ans:
		print('status: {}'.format(ans))


if __name__ == '__main__':
	main()



# #!/usr/bin/env python
#
# ##############################################
# # The MIT License (MIT)
# # Copyright (c) 2016 Kevin Walchko
# # see LICENSE for full details
# ##############################################
# # ----------------------------
# # Simple tool to change the id number of a servo
# #
#
# from __future__ import print_function, division
# from pyxl320 import ServoSerial, Packet, xl320
# # from pyxl320 import DummySerial
# import argparse
# import sys
# PY3 = sys.version_info > (3,)
#
#
# def get_input(s):
# 	"""Handle difference between py2 and py3"""
# 	if PY3:
# 		return input(s)
# 	else:
# 		return raw_input(s)
#
#
# def makeServoIDPacket(curr_id, new_id):
# 	"""
# 	Given the current ID, returns a packet to set the servo to a new ID
# 	"""
# 	pkt = Packet.makeWritePacket(curr_id, xl320.XL320_ID, [new_id])
# 	return pkt
#
#
# def makeInterActive():
# 	port = get_input('Enter serial port >> ')
# 	curr_id = get_input('Enter current id >> ')
# 	new_id = get_input('Enter new id >> ')
# 	return port, curr_id, new_id
#
#
# def handleArgs():
# 	parser = argparse.ArgumentParser(description='set servo id')
# 	parser.add_argument('-i', '--interactive', help='input via commandline', action='store_true')
# 	parser.add_argument('-n', '--new_id', help='set new id', type=int, default=1)
# 	parser.add_argument('-c', '--current_id', help='current id', type=int, default=1)
# 	parser.add_argument('port', help='serial port', type=str)
#
# 	args = vars(parser.parse_args())
# 	return args
#
#
# def main():
# 	args = handleArgs()
#
# 	if args['interactive']:
# 		port, curr_id, new_id = makeInterActive()
# 	else:
# 		port = args['port']
# 		curr_id = args['current_id']
# 		new_id = args['new_id']
#
# 	if port == 'dummy':
# 		ser = ServoSerial(port=port, fake=True)
# 	else:
# 		ser = ServoSerial(port=port)
# 	ser.open()
#
# 	pkt = makeServoIDPacket(curr_id, new_id)
# 	print('pkt', pkt)
#
# 	err, err_str = ser.sendPkt(pkt)
# 	if err:
# 		print('Error: {} {}'.format(err, err_str))
#
#
# if __name__ == '__main__':
# 	main()



###############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import division
from __future__ import print_function
from .ax12 import AX12
from .xl320 import XL320
from .utils import angle2int, le


# def le(h):
# 	"""
# 	Little-endian, takes a 16b number and returns an array arrange in little
# 	endian or [low_byte, high_byte].
# 	"""
# 	h &= 0xffff  # make sure it is 16 bits
# 	return [h & 0xff, h >> 8]
#
#
# def angle2int(angle, degrees=True):
# 	if degrees:
# 		angle = le(int(angle/300*1023))  # degrees
# 	else:
# 		angle = le(int(angle/5.23598776*1023))  # radians
# 	return angle


# PacketManager
# PacketSystem
# Packet
class Packet(object):
	"""
	This is a wrapper class for the xl-320 and ax-12 servos. It can only talk
	to one servo type at a time.
	"""

	def __init__(self, kind):
		self.base = kind()

	def makePingPacket(self, ID=None):
		"""
		Pings a servo
		"""
		if not ID:
			ID = self.base.BROADCAST_ADDR
		pkt = self.base.makePacket(ID, self.base.PING)
		return pkt

	def makeWritePacket(self, ID, reg, values=None):
		"""
		Creates a packet that writes a value(s) to servo ID at location reg. Make
		sure the values are in little endian (use Packet.le() if necessary) for 16 b
		(word size) values.
		"""
		if values:
			if self.base.SERVO_ID == XL320.SERVO_ID:
				params = le(reg) + values
			else:
				params = [reg] + values
		else:
			if self.base.SERVO_ID == XL320.SERVO_ID:
				params = le(reg)
			else:
				params = [reg]

		pkt = self.base.makePacket(ID, self.base.WRITE, params)
		return pkt

	def makeReadPacket(self, ID, reg, values=None):
		"""
		Creates a packet that reads the register(s) of servo ID at location reg. Make
		sure the values are in little endian (use Packet.le() if necessary) for 16 b
		(word size) values.
		"""
		pkt = self.base.makePacket(ID, self.base.READ, [reg, values])
		return pkt

	def makeResetPacket(self, ID, level=0):
		"""
		Resets a servo.
		"""
		if self.base.SERVO_ID == XL320.SERVO_ID:
			params = [XL320.RESET_ALL_BUT_ID]
		else:
			params = None

		pkt = self.base.makePacket(ID, self.base.RESET, params)
		return pkt

	def makeRebootPacket(self, ID):
		"""
		Reboots a servo
		"""

		pkt = self.base.makePacket(ID, self.base.REBOOT)
		return pkt

	def makeServoMovePacket(self, ID, angle, degrees=True):
		"""
		Commands the servo to an angle (in degrees)
		"""
		# if degrees and not (0.0 <= angle <= 300.0):
		# 	raise Exception('makeServoMovePacket(), angle [deg] out of bounds: {}'.format(angle))
		# elif (not degrees) and (not (0.0 <= angle <= 5.23598776)):
		# 	raise Exception('makeServoMovePacket(), angle [rads] out of bounds: {}'.format(angle))
		# val = int(angle/300*1023)
		val = angle2int(angle, degrees)

		pkt = self.makeWritePacket(ID, self.base.GOAL_POSITION, val)
		return pkt

	def makeSyncMovePacket(self, info, degrees=True):
		"""
		Write sync angle information to servos.

		info = [[ID, angle], [ID, angle], ...]
		ID: servo ID
		angle: 0-300 degrees or in radians
		"""
		data = []

		# since all servo angles have the same register addr (GOAL_POSITION)
		# and data size (2), a sinc packet is smart choice
		# compare bulk vs sync for the same commands:
		# bulk = 94 bytes
		# sync = 50 bytes
		for cmd in info:
			data.append(cmd[0])  # ID
			angle = angle2int(cmd[1], degrees)
			data.append(angle[0])  # LSB
			data.append(angle[1])  # MSB

		pkt = self.makeSyncWritePacket(self.base.GOAL_POSITION, data)
		return pkt

	def makeSyncWritePacket(self, reg, info):
		"""
		Write sync angle information to servos.

		info = [[ID, data1, ...], [ID, data1, ...], ...]
		"""
		data = []
		data.append(reg)  # addr
		# length = (len(info[0])+1)*(len(info)+4)
		# data.append(length)  # data length
		data.append(len(info[0])-1)
		for cmd in info:
			data += cmd

		ID = self.base.BROADCAST_ADDR
		instr = self.base.SYNC_WRITE
		pkt = self.base.makePacket(ID, instr, data)  # create packet
		return pkt

	def makeBulkReadPacket(self, data):
		"""
		data = [[data len, ID, addr], [data len, ID, addr], ...]
		"""
		ID = self.base.BROADCAST_ADDR
		instr = self.base.BULK_READ
		pkt = self.base.makePacket(ID, instr, data)  # create packet
		return pkt

	def makeLEDPacket(self, ID, value):
		"""
		Turn on/off the servo LED and also sets the color.
		"""
		if self.base.SERVO_ID == AX12.SERVO_ID:
			if value < 0 or value > 1:
				raise Exception("ERROR: LED can only be off or on: {}".format(value))
			# value = [value]
		# elif self.base.SERVO_ID == XL320.SERVO_ID:
		# 	# print('Not implemented yet')
		# 	value = [0, value]

		pkt = self.makeWritePacket(ID, self.base.LED, [value])
		return pkt

	def decodePacket(self, pkts):
		return self.base.find_packets(pkts)

	def processStatusPacket(self, pkt):
		return self.base.status_packet(pkt)



# ##############################################
# # The MIT License (MIT)
# # Copyright (c) 2016 Kevin Walchko
# # see LICENSE for full details
# ##############################################
#


from __future__ import division
from __future__ import print_function
# from .utils import le, angle2int


class AX12(object):
	"""
	This class handles the AX-12A servo using Dynimel's Protocol version 1.0.
	"""
	SERVO_ID = 1

	# --------- INSTRUCTIONS -----
	PING      = 0x01
	READ      = 0x02
	WRITE     = 0x03
	REG_WRITE = 0x04
	ACTION    = 0x05
	RESET     = 0x06
	# REBOOT    = 0x08
	# STATUS    = 0x55
	# SYNC_READ  = 0x82
	SYNC_WRITE = 0x83
	BULK_READ  = 0x92
	# BULK_WRITE = 0x93

	# -------- EEPROM -------------
	MODEL_NUMBER    = 0
	VER_FIRMWARE    = 2
	ID              = 3
	BAUD_RATE       = 4
	DELAY_TIME      = 5
	CW_ANGLE_LIMIT  = 6   # min angle, default 0
	CCW_ANGLE_LIMIT = 8   # max angle, default 300
	CONTROL_MODE    = 11  # joint or wheel mode, default joint (servo)
	MAX_TORQUE      = 15
	RETURN_LEVEL    = 17

	# -------- RAM ----------------
	TORQUE_ENABLE    = 24  # servo mode on/off - turn into wheel
	LED              = 25
	GOAL_POSITION    = 30
	GOAL_VELOCITY    = 32
	GOAL_TORQUE      = 35
	PRESENT_POSITION = 37  # current servo angle
	PRESENT_SPEED    = 39  # current speed
	PESENT_LOAD      = 41  # current load
	PESENT_VOLTAGE   = 45  # current voltage
	PESENT_TEMP      = 46  # current temperature
	MOVING           = 49
	HW_ERROR_STATUS  = 50
	PUNCH            = 51

	# --------- OTHER -------------
	RESET_ALL                  = 0xFF
	RESET_ALL_BUT_ID           = 0x01
	RESET_ALL_BUT_ID_BAUD_RATE = 0x02
	LED_ON                     = 1
	LED_OFF                    = 0
	BROADCAST_ADDR             = 0xfe  # a packet with this ID will go to all servos
	WHEEL_MODE                 = 1
	JOINT_MODE                 = 2  # normal servo
	DR_1000000                 = 1  # bps = 2000000/(data + 1)

	def __init__(self):
		pass

	@staticmethod
	def check_sum(data):
		"""
		checksum = 255 - ((id + length + data1 + data2 + ... + dataN) & 255)
		"""
		# print(data)
		return 255 - (sum(data) & 255)

	def makePacket(self, ID, instr, params=None):
		"""
		This makes a generic packet.

		0xFF, 0xFF, ID, LENGTH, INST, PARAM 1, PARAM 2, ..., PARAM N, CHECKSUM]
		in:
			ID - servo id
			instr - instruction
			params - instruction parameter values
		out: packet
		"""
		pkt = [0xff, 0xff, ID]
		if params:
			# print(params)
			length = len(params) + 2
			pkt += [length, instr]
			pkt += params
			# print(self.check_sum(pkt[2:]))
			pkt += [self.check_sum(pkt[2:])]
		else:
			pkt += [2, instr, self.check_sum([ID, 2, instr])]

		return pkt

	def find_packets(self, pkt):
		"""
		Search through a string of binary for a valid ax-12 package.

		in: buffer to search through
		out: a list of valid data packet
		"""
		# print('findpkt', pkt)
		# print('-----------------------')
		ret = []
		while len(pkt)-6 >= 0:
			try:
				# print(pkt)
				# check for header
				if pkt[0:2] != [0xFF, 0xFF]:
					pkt.pop(0)  # get rid of the first index
					continue
				length = pkt[3]  # get length
				crc_pos = 3 + length
				pkt_crc = pkt[crc_pos]  # packet crc
				crc = self.check_sum(pkt[2:crc_pos])  # calculated crc
				if pkt_crc == crc:
					pkt_end = crc_pos+1
					ret.append(pkt[:pkt_end])
					del pkt[:pkt_end]
				else:
					pkt.pop(0)
			except Exception:
				pkt.pop(0)
		return ret

	def status_packet(self, pkt):
		def getError(err):
			errors = [
				'Input Voltage',  # 0
				'Angle Limit',
				'Overheating',
				'Range',
				'Checksum',
				'Overload',
				'Instrunction',
				'None'  # 7
			]
			# ret = None
			if not (err == 128 or err == 0):
				err_str = []
				for i in range(0, 8):
					if (err >> i) & 1:
						err_str.append(errors[i])
				ret = ','.join(err_str)
			else:
				ret = errors[7]
			return ret

		ret = {
			'id': pkt[2],
			'error str': getError(pkt[4]),
			'error num': pkt[4],
			'params': pkt[5:-1],
			'raw': list(pkt)
		}
		return ret



# #!/usr/bin/env python
#
# ##############################################
# # The MIT License (MIT)
# # Copyright (c) 2016 Kevin Walchko
# # see LICENSE for full details
# ##############################################
#
# from __future__ import print_function, division
# from pyxl320 import ServoSerial
# from pyxl320 import Packet
# # from pyxl320 import DummySerial
# import argparse
#
# rates = {
# 	9600: 0,
# 	57600: 1,
# 	115200: 2,
# 	1000000: 3
# }
#
#
# def handleArgs():
# 	parser = argparse.ArgumentParser(description='set servo buad rate')
# 	# parser.add_argument('-n', '--newid', help='set new id', type=int, default=1)
# 	parser.add_argument('-c', '--current_rate', help='current baud rate, default is 1000000', type=int, default=1000000)
# 	parser.add_argument('-i', '--id', help='current id', type=int, default=1)
# 	parser.add_argument('port', help='serial port', type=str)
# 	parser.add_argument('-r', '--new_rate', help='set new baud rate to 9600, 57600, 115200, or 1000000; default is 1000000', type=int, default=1000000)
#
# 	args = vars(parser.parse_args())
# 	return args
#
#
# def main():
# 	args = handleArgs()
#
# 	port = args['port']
# 	ID = args['id']
# 	curr_rate = args['current_rate']
# 	# new_id = args['newid']
# 	# 0: 9600, 1:57600, 2:115200, 3:1Mbps
# 	new_rate = rates[args['new_rate']]
#
# 	if port == 'dummy':
# 		ser = ServoSerial(port=port, baud_rate=curr_rate, fake=True)
# 	else:
# 		ser = ServoSerial(port=port, baud_rate=curr_rate)
# 	ser.open()
#
# 	pkt = Packet.makeBaudRatePacket(ID, new_rate)
# 	ser.sendPkt(pkt)
#
# 	# pkt = Packet.makeRebootPacket(ID)
# 	# ser.sendPkt(pkt)
#
# if __name__ == '__main__':
# 	main()



from .servoserial import ServoSerial
from .packet import Packet
from .ax12 import AX12
from .xl320 import XL320
from .packet import Packet


__copyright__ = 'Copyright (c) 2016 Kevin Walchko'
__license__ = 'MIT'
__author__ = 'Kevin J. Walchko'
__version__ = '1.0.2'



# #!/usr/bin/env python
#
# ##############################################
# # The MIT License (MIT)
# # Copyright (c) 2017 Kevin Walchko
# # see LICENSE for full details
# ##############################################
#
# from __future__ import print_function, division
# from pyxl320 import ServoSerial
# from pyxl320.Packet import le, makeReadPacket
# import argparse
# import simplejson as json
# # from pyxl320.xl320 import ErrorStatusMsg
#
#
# # def makeSyncReadPacket(reg, length, ids):
# # 	"""
# # 	Write sync angle information to servos.
# #
# # 	info = [[ID, angle], [ID, angle], ...]
# # 	"""
# # 	# length = le(length)
# # 	data = []
# #
# # 	d = le(reg)
# # 	data.append(d[0])
# # 	data.append(d[1])
# # 	d = le(length)
# # 	data.append(d[0])
# # 	data.append(d[1])
# # 	data += ids
# #
# # 	ID = xl320.XL320_BROADCAST_ADDR
# # 	instr = xl320.XL320_SYNC_READ
# # 	pkt = makePacket(ID, instr, None, data)  # create packet
# #
# # 	# print(pkt)
# #
# # 	return pkt
#
#
# def writeToFile(data, filename='data.json'):
# 	with open(filename, 'w') as outfile:
# 		json.dump(data, outfile)
#
#
# DESCRIPTION = """
# Returns the angles of servos between <start> and <end>. The servo search range
# has to be continous. You can not select servos 1, 3, 6, 7; you have to select
# 1 - 7.
#
# example: get_angle.py /dev/serial0 1 7
#
# Servos: 1 - 7
#    # | Angle
# ----------------------------------------
#    1 | 143.70
#    2 | 281.23
#    3 |  97.07
#    4 | 147.51
#    5 | 280.06
#    6 | 105.87
#    7 | 154.84
# """
#
#
# def handleArgs():
# 	parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
# 	parser.add_argument('port', help='serial port  or \'dummy\' for testing', type=str)
# 	parser.add_argument('start', help='starting servo number', type=int)
# 	parser.add_argument('end', help='ending servo number', type=int)
# 	parser.add_argument('-j', '--json', help='save info to a json file, you must supply a file name: --json my_file.json', type=str)
#
# 	args = vars(parser.parse_args())
# 	return args
#
#
# def getInfo(pkt):
# 	ID = pkt[4]
# 	angle = float((pkt[10] << 8) + pkt[9])/1023
# 	angle *= 300.0
# 	return ID, angle
#
#
# def getSingle(ID, ser):
# 	pkt = makeReadPacket(ID, 37, le(2))
# 	# print('made packet:', pkt)
# 	ID = None
# 	angle = None
#
# 	ans = ser.sendPkt(pkt)
# 	if ans:
# 		ans = ans[0]
# 		ID, angle = getInfo(ans)
#
# 	return ID, angle
#
#
# def main():
# 	args = handleArgs()
# 	port = args['port']
#
# 	if port.lower() == 'dummy':
# 		s = ServoSerial(port=port, fake=True)
# 	else:
# 		s = ServoSerial(port=port)
# 	s.open()
#
# 	ids = range(args['start'], args['end'] + 1)
#
# 	resp = {}
# 	for k in ids:
# 		resp[k] = None
#
# 	# as more servos add up, I might need to increase the cnt number???
# 	for i in ids:
# 		ID, angle = getSingle(i, s)
# 		resp[i] = angle
#
# 	for _ in range(3):
# 		cnt = 0
# 		for k, v in resp.items():
# 			# search through and find servos w/o responses (i.e., None)
# 			if v is None:
# 				cnt += 1  # found a None
# 				ID, angle = getSingle(k, s)
# 				resp[k] = angle
#
# 		if cnt == 0:  # we must have gotten them all on the last run
# 			break
#
# 	print('')
# 	print('Servos: {} - {}'.format(args['start'], args['end']))
# 	print('{:>4} | {:6}'.format('#', 'Angle'))
# 	print('-' * 40)
# 	for k, v in resp.items():
# 		if v is None:
# 			print('{:4} | {}'.format(k, 'unknown'))
# 		else:
# 			print('{:4} | {:6.2f}'.format(k, v))
# 	print('')
#
# 	if args['json']:
# 		print('Saving servo angle info to {}'.format(args['json']))
# 		writeToFile(resp, args['json'])
#
# 	s.close()
#
#
# if __name__ == '__main__':
# 	main()
